module vga_driver(clk, reset,
		  h_sync, v_sync, h_rgb, v_rgb,
		  h_pos, v_pos,
		  kbd_data, kbd_int,
		  h_sync_out, v_sync_out, rgb_out);
   input clk, reset;
   input h_sync, v_sync, h_rgb, v_rgb;
   input [7:0] kbd_data;
   input       kbd_int;
   output      h_sync_out, v_sync_out;
   output [5:0] rgb_out;
   output [9:0] h_pos, v_pos;


   // Sync signal pipeline/delay
   // NOTE: make sure this is *as deep* as the pixel value pipeline.
   reg [1:0] 	hv_sync0, hv_sync1, hv_sync2;
   assign h_sync_out = hv_sync2[1];
   assign v_sync_out = hv_sync2[0];

   always @(posedge clk) begin
      hv_sync0 <= {h_sync, v_sync};
      hv_sync1 <= hv_sync0;
      hv_sync2 <= hv_sync1;
   end // always @ (posedge clk)


   // Screen position counter functionality
   reg [9:0] 	h_pos, v_pos;
   reg [1:0] 	prev_hv_rgb;
   wire [1:0] 	hv_rgb;
   wire [1:0] 	hv_blank;

   parameter 	HV_VBLANK = 2'b00, HV_HBLANK = 2'b01, HV_ACTIVE = 2'b11;

   assign hv_rgb = {h_rgb, v_rgb};
   // hv_blank is for texture units below and operates on t-1 time.
   assign hv_blank = { (prev_hv_rgb == HV_HBLANK),
		       (prev_hv_rgb == HV_VBLANK) };

   // horizontal and vertical position drivers   
   always @(posedge clk) begin
      case(hv_rgb)
	HV_VBLANK: begin
	   h_pos <= 0;
	   v_pos <= -1;
	end // case: HV_VBLANK

	HV_ACTIVE: begin
	   if(prev_hv_rgb != HV_ACTIVE) begin
  	      h_pos <= h_pos + 1;
	      $display("display active, x=%h y=%h", h_pos, v_pos);
	   end else begin
	      h_pos <= 0;
	      v_pos <= v_pos + 1;
	   end
	end // case: HV_ACTIVE

      endcase // case(hvstate)
      prev_hv_rgb <= hv_rgb;
   end


   // Game control unit
   // Self-contains the game state registers, as well as the game
   // control logic.

   // Access to maze RAM
   wire [7:0] maccess_addr;
   wire [3:0] maccess_value, maccess_we, maccess_data;

   // color registers
   wire [5:0] maze_bg, maze_fg;

   vga_game_control local_gamecontrol(clk, reset, hv_blank,
				      kbd_data, kbd_irq,
				      maccess_addr, maccess_value,
				      maccess_data, maccess_we,
				      maze_bg, maze_fg);

   // Maze display unit
   wire       display_bit;
   reg [5:0]  maze_rgb_out;

   maze_display_unit local_maze_display(clk, h_pos, v_pos,
					maccess_addr[7:4], maccess_addr[3:0], maccess_value,
					maccess_we, maccess_data,
					display_bit);

   // RGB output
   always @(posedge clk) begin
      maze_rgb_out <= display_bit ? maze_fg : maze_bg;
   end // always @ (posedge clk)

   
   assign rgb_out = maze_rgb_out;
   //reg [5:0] 	rgb_out;
   
endmodule // vga_driver


module vga_game_control(clk, reset, hv_blank,
			kbd_data, kbd_irq,
			maccess_addr, maccess_read, maccess_write, maccess_we,
			maze_bg, maze_fg);

   input clk, reset;
   input [1:0] hv_blank;
   input [7:0] kbd_data;
   input       kbd_irq;

   output [7:0] maccess_addr;
   input [3:0] maccess_read;
   output [3:0] maccess_write, maccess_we;

   output [5:0] maze_bg, maze_fg;


   reg [1:0] 	last_hv_blank;
   wire 	positive_vblank = (hv_blank[0] && !last_hv_blank[0]);

   // Maze memory registers
   reg [7:0] 	maccess_addr;
   reg [3:0] 	maccess_write, maccess_we;

   // Game state registers live in the context of this module
   reg [5:0] 	maze_bg, maze_fg;


   always @(posedge clk) begin
      
      if(reset) begin
	 maze_bg <= 6'b000000;
	 maze_fg <= 6'b111111;
	 maccess_addr <= 0;
	 maccess_we <= 4'b0000;
	 maccess_write <= 0;

      end else begin // if (reset)

	 if(positive_vblank) begin
	    $display("Throw a VBLANK IRQ!");
	    
	 end // if (positive_vblank)

      end // else: !if(reset)

      last_hv_blank <= hv_blank;
   end // always @ (posedge clk)

endmodule


// maze_display_unit
//
// This unit contains a mazekeeper unit, (see mazekeeper.v) which acts
// as its video memory, and a tile ROM with an image for each possible
// cell state of the maze.
// The tile ROM is 1 BPP, as is its display_bit output.  Ideally this should
// look up a color in a palette.
//
// The size given to the maze unit is 16x16, meaning there are four row bits
// and four column bits.  This fits perfectly with an 8-bit microprocessor core.
//
// Access to the maze RAM is provided through the maccess_ lines.
//
// maze_display_unit requires one clock, but likely has relatively long setup
// and propagation times.

module maze_display_unit(clk, h_pos, v_pos,
			 maccess_row, maccess_col, maccess_value,
			 maccess_we, maccess_data,
			 display_bit);

   input     clk;
   input [9:0] h_pos, v_pos;

   input [3:0] maccess_row, maccess_col;
   output [3:0] maccess_value;
   input [3:0] maccess_we, maccess_data;
   output      display_bit;
   
   reg [4:0] 		tile_x, tile_y;
   wire [3:0] 		tileidx; 		
   wire 		x_range;
   wire [5:0] 		x_coord;

   assign x_range = (h_pos[9:6] == 0) ? 0 :
		    ((h_pos[9:5] >= 18) ? 0 : 1);
   assign x_coord = (h_pos[9:5] - 2);

   // Instantiation of the maze keeper
   // Wire the maccess lines directly into it.
   // 20 wide, 15 high, 0:19=10100=5 bits, 0:14=1110=4 bits.
   mazekeeper #(16, 15, 4, 4)
   	mk(clk, v_pos[8:5], x_coord[3:0], tileidx,
	 maccess_row, maccess_col, maccess_value, maccess_we, maccess_data);

   // NOTE: mazekeeper latches tileidx, maccess_value.
   
   // The tile ROM is organized in a 4x4 grid of tiles of size 32x32.
   // (=16 kbits)  KEEP IT SMALL PLZ!
   // The 2 MSBs of the X grid coordinate represent left,right connectivity,
   // The 2 MSBs of the Y grid coordinate represent top,bottom connectivity.
   wire [6:0] 		hack_tile_x, hack_tile_y;
   wire 		tile_bit;
   
   assign hack_tile_x = { tileidx[1], tileidx[3], tile_x };
   assign hack_tile_y = { tileidx[0], tileidx[2], tile_y };

   rom_mazetiles local_tiles(hack_tile_x, hack_tile_y, tile_bit);

   assign display_bit = (x_range & tile_bit);
   
   always @(posedge clk) begin
      tile_x <= h_pos[4:0];
      tile_y <= v_pos[4:0];
   end

endmodule // maze_display_unit

module rom_mazetiles(x,y,bit);
	input [6:0] x, y;
	output bit;
	assign bit = 1;
endmodule

`include "mazekeeper.v"
